#!/usr/bin/env python3
from oslo_log import log
from tveapi.common import resource as res
import falcon
import json
from tveapi.common import utils
from tveapi.common import resource as res

LOG = log.getLogger(__name__)

class Helm(object):
    """
    API for Helm interaction.

    Get TOKEN: TOKEN=$(curl -s -k -d '{"user":"OS_USERNAME","password":"OS_PASSWORD"}' -H "Content-type: application/json" -X GET https://127.0.0.1:17374/v3/auth/tokens)

    API tests:

    list all deployed helm charts : curl -s -H "X-Auth-Token: $TOKEN" --insecure -X GET  https://127.0.0.1:7373/v1/kubernetes/helm/list
    get status for specific helm deployment : curl -s -H "X-Auth-Token: $TOKEN" -k --insecure -d '{"name": "happy-panda-1"}' -X GET  https://127.0.0.1:7373/v1/kubernetes/helm/status
    """
    class List(res.BaseResource):
        def on_get(self, req, resp):
            code, stdout, stderr = utils.execute_shell_command('sudo helm list -A --kubeconfig /root/.kube/config -o json')
            LOG.info("helm.py: code {} stdout {} stderr {} ".format(code, stdout, stderr))
            if code != 0:
                self.set_status_400(resp, stderr, stdout)
            else:
                resp.status = falcon.HTTP_200
                resp.body = json.loads(json.dumps(stdout))
            return

        def on_put(self, req, resp):
            self.set_status_400(resp, "Only GET supported.")
            return

    class Status(res.BaseResource):
        def on_get(self, req, resp):
            json_req = self.json_body(req)

            if 'name' not in json_req or json_req['name'] == '':
                self.on_fail(resp, 'Name parameter in jsonData missing.')
                return

            params = ''
            if 'namespace' in json_req:
                params += '-n ' + json_req['namespace']

            code, stdout, stderr = utils.execute_shell_command('sudo helm status {} --show-resources -o json {}'.format(json_req['name'], params))
            LOG.info("helm.py: code {} stdout {} stderr {} ".format(code, stdout, stderr))
            if code != 0:
                self.set_status_400(resp, stderr, stdout)
            else:
                resp.status = falcon.HTTP_200
                resp.body = json.loads(stdout)
            return

        def on_put(self, req, resp):
            self.set_status_400(resp, "Only GET supported.")
            return
