#!/usr/bin/env python3
from collections import defaultdict
from oslo_log import log
from tveapi.common import resource as res
import falcon
import json
import subprocess
import os
import time
from kubernetes import client, config

from tveapi.common import utils
from tveapi.common import resource as res
from tveapi.common import validate_docs as docs
from tveapi import policy

LOG = log.getLogger(__name__)

# curl -H "X-AUTH-TOKEN:$TOKEN" -k --insecure -X GET https://<TVE>:7373/v1/kubernetes/token
# curl -s -H "X-Auth-Token: $TOKEN" --insecure -X GET https://127.0.0.1:7373/v1/kubernetes/worker_status?name=<NODE_NAME>

class TVE(object):
    class token(res.BaseResource):
        def on_get(self, req, resp):
            code, stdout, stderr = utils.execute_shell_command('sudo kubeadm token create --print-join-command --groups "system:bootstrappers:kubeadm:default-node-token" --ttl=1h')
            LOG.info("kubernetes.py: code {} stdout {} stderr {} ".format(code, stdout, stderr))

            if code != 0:
                self.set_status_500(resp, stderr, stdout)
            elif res.SYSTEM_BUSY_MSG in str(stdout):
                self.set_status_system_busy(resp)
            else:
                resp.status = falcon.HTTP_200
                token = stdout.split("--token")[1].split()[0]
                cert = stdout.split("--discovery-token-ca-cert-hash")[1]
                resp.body = "%s %s" % (token,cert)
                #LOG.info(" token {}".format(token))
                #LOG.info(" cert {}".format(cert))
                #LOG.info(" body {}".format(resp.body))
            return

        def on_put(self, req, resp):
            self.set_status_400(resp, "Only GET supported.")
            return

    class worker_status(res.BaseResource):
        def on_get(self, req, resp):
            name = req.get_param('name')
            command = 'sudo python3 /scripts/isNodeReady.py %s' % name
            code, stdout, stderr = utils.execute_shell_command(command)
            LOG.error("kubernetes.py: name {} code {} stdout {} stderr {}".format(name, code, stdout, stderr))

            if code != 0:
                self.set_status_500(resp, stderr, stdout)
            elif res.SYSTEM_BUSY_MSG in str(stdout):
                self.set_status_system_busy(resp)
            else:
                resp.status = falcon.HTTP_200
                resp.body = stdout
            return

        def on_put(self, req, resp):
            self.set_status_400(resp, "Only GET supported.")
            return

    class worker_status_detailed(res.BaseResource):
        def on_get(self, req, resp):
            name=req.get_param('name')
            command = 'sudo python3 /scripts/isNodeReady.py %s node-condition' % name
            code, stdout, stderr = utils.execute_shell_command(command)
            LOG.error("kubernetes.py: name {} code {} stdout {} stderr {}".format(name, code, stdout, stderr))

            if code != 0:
                self.set_status_500(resp, stderr, stdout)
            elif res.SYSTEM_BUSY_MSG in str(stdout):
                self.set_status_system_busy(resp)
            else:
                resp.status = falcon.HTTP_200
                resp.body = {"status": "{}".format(stdout)}
            return

    class worker_status_based_on_pods(res.BaseResource):

        def on_get(self, req, resp):
            name = req.get_param('name')
            command = 'sudo python3 /scripts/isNodeReady.py %s kube-system-pods' % name
            code, stdout, stderr = utils.execute_shell_command(command)
            LOG.error("kubernetes.py: name {} code {} stdout {} stderr {}".format(name, code, stdout, stderr))

            if code != 0:
                self.set_status_500(resp, stderr, stdout)
            elif res.SYSTEM_BUSY_MSG in str(stdout):
                self.set_status_system_busy(resp)
            else:
                resp.status = falcon.HTTP_200
                resp.body = stdout
            return

        def on_put(self, req, resp):
            self.set_status_400(resp, "Only GET supported.")
            return

    @falcon.before(res.BaseResource.k8s_control_plane_eligible())
    class worker_clean(res.BaseResource):

        def on_put(self, req, resp):
            workers = self.json_body(req)
            drain_command = "export KUBECONFIG=/home/tveapi/.kube/config; kubectl drain {0} --delete-emptydir-data --force --ignore-daemonsets --disable-eviction --grace-period=0"
            delete_command = "export KUBECONFIG=/home/tveapi/.kube/config; kubectl delete node {0}"
            results = {}
            code_to_msg = defaultdict(str)
            for worker in workers:
                for command in (drain_command, delete_command):
                    code, stdout, stderr = utils.execute_shell_command(command.format(worker))
                    msg = f"{stdout}{stderr if code else ''}"
                    if code_to_msg['1' if code else '0'] != msg:
                        code_to_msg['1' if code else '0'] += msg
                    results[worker] = dict(code_to_msg)
                code_to_msg.clear()

            for worker in workers:                       ## If either one of the 
                if results[worker].keys() >= {'0', '1'}: ## commands failed 
                    results[worker].pop('0')             ## report the failed one only

            resp.status = falcon.HTTP_200
            resp.body = json.dumps(results)

            return

        def on_get(self, req, resp):
            self.set_status_400(resp, "Only PUT supported.")
            return

    @falcon.before(res.BaseResource.k8s_control_plane_eligible())
    class worker_nodes(res.BaseResource):
        def on_get(self, req, resp):
            command = "export KUBECONFIG=/home/tveapi/.kube/config; kubectl get nodes -A -l node-role.kubernetes.io/control-plane!=\"\" -o jsonpath=\"{range .items[*]}{.metadata.name}{'\t'}\""
            code, stdout, stderr = utils.execute_shell_command(command)
            
            if code != 0:
                self.set_status_500(resp, stderr, stdout)
            elif res.SYSTEM_BUSY_MSG in str(stdout):
                self.set_status_system_busy(resp)
            else:
                nodes = stdout.split()
                
                resp.status = falcon.HTTP_200
                resp.body = json.dumps({'result': nodes})
            return

    # Not using
    class k8s_token (res.BaseResource):
        def on_get(self, req, resp):
            command = "export KUBECONFIG=/home/tveapi/.kube/config; kubectl get secrets -o jsonpath=\" {.items[?(@.metadata.annotations['kubernetes\.io/service-account\.name']=='default')].data.token} \""
            code, stdout, stderr = utils.execute_shell_command(command)
            #LOG.error("kubernetes.py: code {} stdout {} stderr {} ".format(code, stdout, stderr))

            if code != 0:
                self.set_status_500(resp, stderr, stdout)
                LOG.error("kubernetes.py: code {} stdout {} stderr {} ".format(code, stdout, stderr))
            elif res.SYSTEM_BUSY_MSG in str(stdout):
                self.set_status_system_busy(resp)
            else:
                command = 'echo %s | base64 --decode' % stdout
                token_code, token_stdout, token_stderr = utils.execute_shell_command(command)
                #LOG.error("kubernetes.py: code {} stderr {} ".format(token_code, token_stderr))
                #LOG.error("kubernetes.py: stdout {}".format(token_stdout))
                resp.status = falcon.HTTP_200
                resp.body = token_stdout
                #LOG.info(" token {}".format(resp.body))
            return

        def on_put(self, req, resp):
            self.set_status_400(resp, "Only GET supported.")
            return

    class main_vpp_int(res.BaseResource):
        def on_put(self, req, resp):
            json_req = self.json_body(req)
            node = json_req['name']
            int_name = json_req['int_name']
            cmd = json_req['cmd']

            code, stdout, stderr = utils.execute_shell_command('sudo python3 /scripts/set-vpp-main-int.py {} {} {}'.format(node, int_name, cmd))
            if code != 0:
                self.set_status_500(resp, stderr, stdout)
            else:
                resp.status = falcon.HTTP_200
                resp.body = stdout

            return

        def on_get(self, req, resp):
            self.set_status_400(resp, "Only PUT supported.")
            return

    class dhcp(res.BaseResource):
        """
        OS_USERNAME, OS_PASSWORD for the internal system user. See local-openrc.env
        Get TOKEN : TOKEN=$(curl -s -k -d '{"user":"OS_USERNAME","password":"OS_PASSWORD"}' -H "Content-type: application/json" -X GET https://127.0.0.1:17374/v3/auth/tokens)

        API tests:
        create DHCP : curl -s -H "X-Auth-Token: $TOKEN" -H 'Content-Type: application/json' -k -X POST  https://127.0.0.1:7373/v1/kubernetes/dhcp \
                        -d '{ "dhcp": { "dns": "1.2.3.1, 1.2.3.2", "domain": "telco.com", "serverAddress": "1.2.3.6", "gateway": "1.2.3.1", "networkInterface": "nat0", \
                        "range": "1.2.3.10-1.2.3.100", "subnet": "1.2.3.1/24", "staticConfigurations": [ { "fixedIpAddress": "1.2.3.66", \
                        "hardwareEthernetAddress": "da:b7:f5:d8:e2:97", "host": "fedora-vmi" } ] } }'

        delete DHCP : curl -s -H "X-Auth-Token: $TOKEN" -H 'Content-Type: application/json' -k -X DELETE  https://127.0.0.1:7373/v1/kubernetes/dhcp \
                        -d '{ "networkInterface": "nat0" }'
        """

        @policy.enforce('kubernetes.dhcp:post')
        @falcon.before(res.BaseResource.max_body())
        def on_post(self, req, resp):
            doc = self.json_body(req)
            docs.KubernetesDhcpConfigDoc().validate(doc)
            ifname = doc['dhcp']['networkInterface']

            if 'staticConfigurations' not in doc['dhcp']:
                doc['dhcp']['staticConfigurations'] = None

            try:
                if not os.access(res.K8S_DHCP_DIR, os.W_OK):
                    utils.execute_shell_command('sudo chmod 777 {}'.format(res.K8S_DHCP_DIR))

                utils.execute_shell_command(f'sudo ip netns add {ifname}_dhcpd')
                fd = os.open(res.K8S_DHCP_DIR + ifname, os.O_RDWR|os.O_CREAT)
                config = os.fdopen(fd, "w+")
                config.seek(0)
                config.truncate()
                json.dump(doc, config)
                time.sleep(5)
                config.close()
            except Exception as e:
                self.set_status_500(resp, e)
                return

            time.sleep(10)

            code, stdout, stderr = utils.execute_shell_command('/usr/bin/sudo python3 /scripts/kubernetes_dhcp.py {} {} {}'.format(res.K8S_DHCP_DIR, ifname, 'CHECK_STATUS'))
            if code != 0:
                os.remove(res.K8S_DHCP_DIR + ifname)
                self.set_status_500(resp, stderr, stdout)
            else:
                resp.body = {'create': 'ok'}
                resp.status = falcon.HTTP_201
            return

        @policy.enforce('kubernetes.dhcp:delete')
        @falcon.before(res.BaseResource.max_body())
        def on_delete(self, req, resp):
            json_req = self.json_body(req)
            ifname = json_req['networkInterface']

            utils.execute_shell_command(f'sudo ip netns delete {ifname}_dhcpd')

            try:
                if os.path.exists(res.K8S_DHCP_DIR + ifname):
                    os.remove(res.K8S_DHCP_DIR + ifname)

                resp.body = {'delete': 'ok'}
                resp.status = falcon.HTTP_200
            except Exception as e:
                self.set_status_500(resp, 'Configuration for {} is missing!'.format(ifname))
            return

        def on_get(self, req, resp):
            self.set_status_400(resp, "Only PUT supported.")
            return


    class cert(res.BaseResource):
        def on_get(self, req, resp):
            cert_key = '/home/tveapi/join_cert_key'

            try:
                if not os.path.exists(cert_key):
                    with open(cert_key, 'w') as f:
                        f.write('')

                diffs = int(time.time()) - int(os.path.getmtime(cert_key))
                if diffs > 3300 or os.path.getsize(cert_key) != 64:
                    # If cert file is older that 55 minutes, create the new one
                    code, stdout, stderr = utils.execute_shell_command('sudo kubeadm init phase upload-certs --upload-certs')
                    LOG.debug('kubernetes.py: code {} stdout {} stderr {} '.format(code, stdout, stderr))

                    if code != 0:
                        self.set_status_500(resp, stderr, stdout)
                    elif res.SYSTEM_BUSY_MSG in str(stdout):
                        self.set_status_system_busy(resp)
                    else:
                        resp.status = falcon.HTTP_200
                        cert = stdout.split('\n')[2]
                        with open(cert_key, 'w') as f:
                            f.write(cert)
                        resp.body = '{}'.format(cert)
                else:
                    resp.status = falcon.HTTP_200
                    with open(cert_key, 'r') as f:
                        resp.body = f.readline()
            except Exception as e:
                self.set_status_500(resp, 'failed to generate control-plane certificate key')

            return

        def on_put(self, req, resp):
            self.set_status_400(resp, "Only GET supported.")
            return

    class pci_passthrough(res.BaseResource):
        '''
        Update fields of a Kubevirt pciHostDevices resource using a JSON patch. JSON formats are accepted.

        :param operation: The patch operations supported by JSON Patch are 'add', 'replaceall'. Type: string. Default: 'add'.
        :param data: Data in json format. Type: array. Required: ["externalResourceProvider", "pciVendorSelector", "resourceName"].

        API examples
        -------------------
        Get PCI devices:
        curl -s -H "X-Auth-Token: $TOKEN" -H 'Content-Type: application/json' -k -X GET  https://127.0.0.1:7373/v1/kubernetes/pci_passthrough
        Add PCI device:
        curl -s -H "X-Auth-Token: $TOKEN" -H 'Content-Type: application/json' -k -X POST  https://127.0.0.1:7373/v1/kubernetes/pci_passthrough -d \
            '{ "operation": "add", "data": [{"externalResourceProvider": false, "pciVendorSelector": "10de:2534", "resourceName": "vfio/10de_2534"}] }'
        '''

        @policy.enforce('kubernetes.pci_passthrough:post')
        @falcon.before(res.BaseResource.max_body())
        def on_post(self, req, resp):
            json_req = self.json_body(req)
            docs.KubernetesPciPassthroughConfigDoc.validate(json_req)

            if 'operation' not in json_req:
                operation = 'add'
            else:
                operation = json_req['operation']

            try:
                # Kubevirt API init
                config.load_kube_config()
                api = client.CustomObjectsApi()

                # Get all resources
                current_data = api.list_namespaced_custom_object(
                    group='kubevirt.io',
                    version='v1',
                    namespace='kubevirt',
                    plural='kubevirts')

                if operation == 'add':
                    data = json_req['data'][0]
                    devices = current_data['items'][0]['spec']['configuration']['permittedHostDevices']
                    if devices:
                        if data not in devices['pciHostDevices']:
                            devices['pciHostDevices'].append(data)
                            data = devices['pciHostDevices']
                        else:
                            resp.status = falcon.HTTP_200
                            resp.body = {"configured":"ok"}
                            return
                    else:
                        data = json_req['data']
                elif operation == 'replaceall':
                    data = json_req['data']

                # Create patch
                patch = {
                    "spec": {
                        "configuration": {
                            "permittedHostDevices": {
                                "pciHostDevices": data
                            }
                        }
                    }
                }

                # Patch custom resource
                api.patch_namespaced_custom_object(
                        group='kubevirt.io',
                        version='v1',
                        namespace='kubevirt',
                        plural='kubevirts',
                        name='kubevirt',
                        body=patch)

            except KeyError as e:
                self.on_fail(resp, f'Kubevirt PCI passthrough feature is not enabled!')
                return
            except Exception as e:
                self.on_fail(resp, f'Unable to get Kubernetes resources - {e}')
                return

            resp.status = falcon.HTTP_200
            resp.body = {"configured":"ok"}

        def on_get(self, req, resp):
            try:
                # Kubevirt API init
                config.load_kube_config()
                api = client.CustomObjectsApi()

                # Get all resources
                current_data = api.list_namespaced_custom_object(
                    group='kubevirt.io',
                    version='v1',
                    namespace='kubevirt',
                    plural='kubevirts')

                devices = current_data['items'][0]['spec']['configuration']['permittedHostDevices']

                resp.status = falcon.HTTP_200
                resp.body = json.loads(json.dumps(devices))
            except Exception as e:
                self.on_fail(resp, f'Unable to get Kubernetes resources - {e}')
                return
